#! /usr/bin/python3

# Eprom emulator
# Author   : San Bergmans
# Date     : Early 2018
# Web site : www.sbprojects.net

# Use: epromemulator.py [-f] [-h] [-a] [-o OFFSET] Eprom_Type File_Name

import sys
import os
import time
import argparse

try:
    import RPi.GPIO as GPIO
except ImportError:
    print("*** Error: Importing RPi.GPIO!", file=sys.stderr)
    sys.exit(1)

Version = "1.00"

# Disable warnings of other processes possibly using the GPIO
GPIO.setwarnings(False)

# Select I/O pin naming convention
# GPIO.setmode(GPIO.BOARD)
GPIO.setmode(GPIO.BCM)

# Define the outputs
RESET_Out = 17
WRITE_Out = 27
SDA_Out = 22
SCL_Out = 5
LDATA_Out = 6
LADDRL_Out = 13
LADDRH_Out = 19
LDEVSEL_Out = 26

print("EPROM Emulator V{}\n".format(Version))


# -----------------------------------------------------------------------

def Eprint(msg):

    """
    Print an error message to stderr.
    """

    if args.Animate:
        # Print newline if animation is enabled
        print()
    print("*** Error: ", msg, file=sys.stderr)


# -----------------------------------------------------------------------

def ParseOffset(argument):

    """
    Parse the offset parameter.
    May be in different notations:
    decimal, hexH, \$hex, or 0xhex
    Please note that $hex must be escaped or quoted with single quotes.
    """

    hex = ''

    if argument.isdigit():
        # It is a decimal value
        return int(argument)

    if argument.upper()[-1] == 'H':
        # It is hexH format
        hex = argument[:-1]
    elif argument[0] == '$':
        # It is $hex format
        hex = argument[1:]
    elif len(argument) > 2 and argument.upper()[0:2] == '0X':
        # It is 0xhex format
        hex = argument[2:]

    try:
        offset = int(hex, 16)
    except ValueError:
        offset = -1

    return offset


# -----------------------------------------------------------------------

def InitOutputs():

    """
    Initialize outputs.
    """

    outputs = [RESET_Out, WRITE_Out, SDA_Out, SCL_Out, LDATA_Out,
               LADDRL_Out, LADDRH_Out, LDEVSEL_Out]
    GPIO.setup(outputs, GPIO.OUT, initial=GPIO.LOW)


# -----------------------------------------------------------------------

def SerialOut(data, register):

    """
    Send 8 bits of data to the output, MSB first.
    Finally latches the data into the selected register.
    """

    for bit in range(0, 8):
        # Do 8 bits, MSB first
        GPIO.output(SDA_Out, 0x80 & (data << bit))
        # Pulse the clock line
        GPIO.output(SCL_Out, GPIO.HIGH)
        GPIO.output(SCL_Out, GPIO.LOW)

    # Pulse the latch line of the selected register
    GPIO.output(register, GPIO.HIGH)
    GPIO.output(register, GPIO.LOW)


# -----------------------------------------------------------------------

def ResetOn():

    """
    Set reset to active.
    Also enables memory access by Pi
    """

    GPIO.output(RESET_Out, GPIO.HIGH)


# -----------------------------------------------------------------------

def ResetOff():

    """
    Set reset to inactive.
    Also enables memory access by target system.
    """

    GPIO.output(RESET_Out, GPIO.LOW)


# -----------------------------------------------------------------------

def SetEprom(devtype):

    """
    Set the device type.
    User can choose between:
    2716 2732 2764 27128 27256 27512 or
    2k 4k 8k 16k 32k 64k or
    2 4 8 16 32 64
    DevSize is set to last available address of device.
    """

    global Dev_Size

    options = {
        '2716': [0b00000000, 0x07FF],
        '2732': [0b00000001, 0x0FFF],
        '2764': [0b00000011, 0x1FFF],
        '27128': [0b00000111, 0x3FFF],
        '27256': [0b00001111, 0x7FFF],
        '27512': [0b00011111, 0xFFFF],
        '2k': [0b00000000, 0x07FF],
        '4k': [0b00000001, 0x0FFF],
        '8k': [0b00000011, 0x1FFF],
        '16k': [0b00000111, 0x3FFF],
        '32k': [0b00001111, 0x7FFF],
        '64k': [0b00011111, 0xFFFF],
        '2': [0b00000000, 0x07FF],
        '4': [0b00000001, 0x0FFF],
        '8': [0b00000011, 0x1FFF],
        '16': [0b00000111, 0x3FFF],
        '32': [0b00001111, 0x7FFF],
        '64': [0b00011111, 0xFFFF]}

    if devtype not in options:
        Eprint("Eprom type not recognized")
        sys.exit(2)

    # Set the selected device type
    SerialOut(options[devtype][0], LDEVSEL_Out)
    Dev_Size = options[devtype][1]


# -----------------------------------------------------------------------

def OpenFile(filename):

    """
    Try to open the target file.
    Returns file handle
    """

    while True:
        # Repeat this untill no errors found
        try:
            handle = open(filename, 'r', errors='ignore')
        except FileNotFoundError:
            Eprint("File not found.")
            sys.exit(3)
        except IOError:
            Eprint("Could not open the file.")
            sys.exit(3)
        except TypeError:
            # This is usually only temporary during assembly
            time.sleep(2)
        else:
            # No error occurred, OK to proceed
            break

    return handle


# -----------------------------------------------------------------------

def CloseFile(filehandle):

    """
    Close the target file again.
    """

    filehandle.close()


# -----------------------------------------------------------------------

def ParseIntelLine(line):

    """
    Parse an IntelHex line.
    Record types 02, 03, 04 and 05 are ignored.
    Maximum payload length is 32 bytes.
    Everything after the checksum is ignored.

    Returns True if EOF record found, or when an error is found.
    """

    global Errors

    # Remove EOL character and leading and trailing white space
    line = line.strip()

    if len(line) == 0:
        # Ignore empty lines
        return False

    # Pad the line with plenty of spaces to make parsing easier
    # We don't have to worry about the line being too short
    line = line + ' '*74

    if line[0] in ('#', ';'):
        # Ignore lines starting with a comment symbol
        return False

    if line[0] != ":":
        # Line must begin with a colon
        Eprint("Illegal start of formatted line.")
        Errors += 1
        return True

    if line[7:9] in ('02', '03', '04', '05'):
        # These record types are simply ignored
        return False

    if line[0:11] == ":00000001FF":
        # End of file marker found
        return True

    datalength = int(line[1:3], 16)
    checksum_index = (datalength + 4) * 2 + 1
    checksum = line[checksum_index:checksum_index + 2]
    address = int(line[3:7], 16)
    data = line[9:checksum_index]

    # Decode the dataline
    bytes = Decode(datalength, checksum, address, data)
    if len(bytes) == 0:
        # Decoding the line resulted in an error
        return True

    # Upload these bytes to Emulator memory
    Upload(address, bytes)


# -----------------------------------------------------------------------

def Decode(len, sum, addr, data):

    """
    Decode an Intel string.

    Returns a list of N bytes.
    Returns an empty list if an error has occurred.
    """

    global Errors

    checksum = len + (addr & 0xFF) + (addr >> 8)
    index = 0
    bytes = []

    if len <= 0 or len > 32:
        # Illegal length
        Eprint("Illegal line length found.")
        Errors += 1
        return bytes

    while len > 0:
        # Iterate over all data bytes
        decimal = int(data[index:index+2], 16)
        bytes.append(decimal)
        checksum = checksum + decimal
        index += 2
        len -= 1

    checksum = (checksum + int(sum, 16)) & 0xFF

    if checksum != 0:
        # Checksum didn't add up
        Eprint("Checksum error.")
        bytes = []
        Errors += 1
        return bytes

    return bytes


# -----------------------------------------------------------------------

def Upload(address, data):

    """
    Write this line of data to the Emulator memory.
    Ignore bytes which are not within the current memory size.
    """

    global Addr_H, Bytes_Written

    address -= Offset

    for byte in data:
        # Iterate over all data bytes
        if address <= Dev_Size and address >= 0:
            # Address is inside address range
            if Addr_H != (address >> 8):
                # A new address high is needed
                if args.Animate:
                    # Append a dot at every Addr_H change
                    print('.', end='', flush=True)
                Addr_H = address >> 8
                SerialOut(Addr_H, LADDRH_Out)
            # Now send address low and data to the shift registers
            SerialOut(address & 0xFF, LADDRL_Out)
            SerialOut(byte, LDATA_Out)
            # And send a write pulse
            GPIO.output(WRITE_Out, GPIO.HIGH)
            GPIO.output(WRITE_Out, GPIO.LOW)
            address += 1
            Bytes_Written += 1

    return


# -----------------------------------------------------------------------
# Main program
# -----------------------------------------------------------------------

InitOutputs()

# Define and parse the arguments
parser = argparse.ArgumentParser(description="Get EPROM type and file name.")
parser.add_argument('-f', action='store_true', default=False,
                    dest='FollowFlag', help='Follow file changes')
parser.add_argument('-a', action='store_true', default=False,
                    dest='Animate', help='Animation during upload')
parser.add_argument('-o', action='store', default='0', dest='Offset',
                    help='Offset address')
parser.add_argument('Eprom_Type', action='store',
                    help='2716, 2732, 2764, 27128, 27256, 27512')
parser.add_argument('File_Name', action='store')
args = parser.parse_args()

# Set EPROM size by device type
SetEprom(args.Eprom_Type)

# Parse the offset argument
Offset = ParseOffset(args.Offset)
if Offset < 0 or Offset > 0xFFFF:
    # An illegal offset has been given
    Eprint("Illegal offset.")
    sys.exit(4)

# Main program loop
while True:
    # Open the target file
    Obj_File = OpenFile(args.File_Name)

    # Now it's time to stop the target system
    ResetOn()

    print("Uploading", args.File_Name)

    Addr_H = -1         # Make Addr_H invalid
    Bytes_Written = 0   # Number of bytes written
    Errors = 0          # Error flag, when != 0

    # Start reading the file line by line
    while True:
        parse_line = Obj_File.readline()
        if len(parse_line) > 0:
            # Not at the end of the file
            if ParseIntelLine(parse_line):
                # EOF record found, or error encountered
                break
        else:
            # End of file, exit this loop
            break

    if Errors == 0:
        # No errors occurred, start target system again
        ResetOff()
        if args.Animate:
            print()
        print(Bytes_Written, "bytes written to emulator memory.")

    # Close the target tfile
    CloseFile(Obj_File)

    if not args.FollowFlag:
        # We're not going to follow the file
        # So we'll quit
        break

    print("Waiting for newer version of", args.File_Name)

    modify_time = os.path.getmtime(args.File_Name)
    while True:
        # Wait until the file timestamp changes
        if modify_time != os.path.getmtime(args.File_Name):
            # The timestamp has changed
            break
        time.sleep(1)

    print("New file on its way.")
    time.sleep(1)
    while True:
        # Now wait until the new file contains an
        # an End Of File record before proceeding
        Obj_File = OpenFile(args.File_Name)
        while True:
            # Try to find the End Of File Record
            parse_line = Obj_File.readline()
            if len(parse_line) > 0:
                # Not at the end of the file yet
                if ":00000001FF" in parse_line:
                    # An EOF record found, file is complete
                    break
            else:
                # End of file, EOF record not found
                # Close and reopen the file, who knows what
                # has happened to it while it growed.
                CloseFile(Obj_File)
                time.sleep(2)
                Obj_File = OpenFile(args.File_Name)

        # The new file is now complete, stop waiting
        break
